# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Classification async plugin for Raspberry Pi"""

import copy
import asyncio
import json
import uuid
import logging
import os
import io
import time
import numpy as np
import picamera
from threading import Thread
from collections import deque

from foglamp.common import logger
from foglamp.plugins.common import utils
import async_ingest

from edgetpu.classification.engine import ClassificationEngine

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Image classification using Imagenet model',
        'type': 'string',
        'default': 'raspi_classify',
        'readonly': 'true'
    },
    'model_file': {
        'description': 'TFlite model file to use for inference',
        'type': 'string',
        'default': 'mobilenet_v2_1.0_224_quant_edgetpu.tflite',
        'order': '1',
        'displayName': 'TFlite model file'
    },
    'labels_file': {
        'description': 'Labels file used during inference',
        'type': 'string',
        'default': 'imagenet_labels.txt',
        'order': '2',
        'displayName': 'Labels file'
    },
    'asset_name': {
        'description': 'Asset Name',
        'type': 'string',
        'default': 'RaspiImageClassify',
        'order': '3',
        'displayName': 'Asset Name'
    }
}


_LOGGER = logger.setup(__name__, level=logging.INFO)


c_callback = None
c_ingest_ref = None
loop = None
t = None
t2 = None
shutdown_in_progress = None


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'Raspberry Pi classification plugin',
        'version': '1.7.0',
        'mode': 'async',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin with the merged config
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    global shutdown_in_progress
    shutdown_in_progress = False
    data = copy.deepcopy(config)
    return data


def plugin_start(handle):
    """ Start the plugin operation - observe video feed from camera device and classify that using the
        configured ML model
        Args:
            handle: handle returned by the plugin initialisation call
        Returns:
            Sends back the reading object (JSON doc) via C/python async ingest interface when it recognizes
            the video frame content
        Raises:
    """
    global loop, t, t2, shutdown_in_progress
    loop = asyncio.new_event_loop()
    try:
        model = handle['model_file']['value']
        labels = handle['labels_file']['value']
        asset_name = handle['asset_name']['value']

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)

        with open(labels, 'r') as f:
            pairs = (l.strip().split(maxsplit=1) for l in f.readlines())
            labels = dict((int(k), v) for k, v in pairs)

        engine = ClassificationEngine(model)

        def camera_loop():
            with picamera.PiCamera() as camera:
                camera.resolution = (640, 480)
                camera.framerate = 30
                camera.annotate_text_size = 20
                _, width, height, channels = engine.get_input_tensor_shape()
                camera.start_preview()
                try:
                    stream = io.BytesIO()
                    fps = deque(maxlen=20)
                    fps.append(time.time())
                    for foo in camera.capture_continuous(stream,
                                                         format='rgb',
                                                         use_video_port=True,
                                                         resize=(width, height)):
                        stream.truncate()
                        stream.seek(0)
                        input = np.frombuffer(stream.getvalue(), dtype=np.uint8)
                        start_ms = time.time()
                        results = engine.ClassifyWithInputTensor(input, top_k=3)
                        inference_ms = (time.time() - start_ms) * 1000.0
                        fps.append(time.time())
                        fps_ms = len(fps) / (fps[-1] - fps[0])
                        camera.annotate_text = "Inference: %5.2fms FPS: %3.1f" % (inference_ms, fps_ms)
                        objs = []
                        for result in results:
                            camera.annotate_text += "\n%.0f%% %s" % (100 * result[1], labels[result[0]])
                            objs.append({'label': labels[result[0]], 'score': 100 * result[1]})
                        print(camera.annotate_text)

                        reads = {
                            'count': len(results),
                            'objects': objs
                        }
                        data = {
                            'asset': asset_name,
                            'timestamp': utils.local_timestamp(),
                            'key': str(uuid.uuid4()),
                            'readings': reads
                        }

                        global shutdown_in_progress
                        if shutdown_in_progress:
                            break

                        if len(results) > 0:
                            _LOGGER.debug(json.dumps(data))
                            if not shutdown_in_progress:
                                async_ingest.ingest_callback(c_callback, c_ingest_ref, data)

                finally:
                    camera.stop_preview()

        def run():
            global loop
            loop.run_forever()

        t = Thread(target=run)
        t.start()

        t2 = Thread(target=camera_loop)
        t2.start()

    except Exception as e:
        _LOGGER.exception("Raspberry Pi classification plugin failed to start. Details: %s", str(e))
        raise


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    it should be called when the configuration of the plugin is changed during the operation of the south service.
    The new configuration category should be passed.

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    _LOGGER.debug("Old config for Raspberry Pi classification plugin {} \n new config {}".format(handle, new_config))

    # TODO: limit complete plugin restart (shutdown + start) to only the cases where absolutely necessary
    plugin_shutdown(handle)
    new_handle = plugin_init(new_config)
    plugin_start(new_handle)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup

    To be called prior to the south service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
    Raises:
    """
    global loop, t, t2, shutdown_in_progress
    try:
        shutdown_in_progress = True
        time.sleep(1)
        loop.stop()
        loop = None
        t = None
        t2 = None
        _LOGGER.info('Raspberry Pi classification plugin shut down.')
    except Exception as e:
        _LOGGER.exception(str(e))
        raise


def plugin_register_ingest(handle, callback, ingest_ref):
    """ Required plugin interface component to communicate to South C server

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C/Python async ingest interface
        ingest_ref: C opaque object required to passed back to C/Python async ingest interface
    """
    global c_callback, c_ingest_ref
    c_callback = callback
    c_ingest_ref = ingest_ref
